/** TMfactory.java -- Base class for generating the right kind of
 * 		TOS_Msg structures. Handles stub creation for each header type.
 *    Stub message types will contain a factory instance and
 *     can be used as a master to generate other specific types
 *     as extended by specific applications.
 * 	  Extend TOSMsgFactory( int AMtype ) to add application
 *     specific message types.
 *    Extend doMessage() to add application behavior for received
 *     messages. If just printing or logging messages as they
 *     arrive use setLogging( MessageLog l ) to set a logger,
 *     or getMessage() to get the last message received.
 *
 * @author schip
 */
package com.etantdonnes.tinyos;

import java.io.IOException;

import net.tinyos.packet.BuildSource;
import net.tinyos.packet.PacketSource;
import net.tinyos.util.PrintStreamMessenger;

// to get rid of whines about unparmeterized TOS_Msg 
@SuppressWarnings("unchecked")

public class TMfactory extends Thread
{
	static boolean tracemsg = false;	// print hex I/O messages
	static boolean errmsg = true;		// print errors & bailouts
	
	// packet types are defined in TOS_Msg to match those in
	//		net.tinyos.packet.Platform
	//   apparently not defined or used anywhere else...those TOSSERS!
	protected int headerType = -1; // the type of message header 
	protected PacketSource packetSource = null; // message source
	
	/* Create a default thingie based on the MOTECOM envvar type.
	 *  Opens a PacketSource object for use in communication
	 *   with the external TOStype device.
	 *   
	 *   Starts a readThread running to get and dispose messages.
	 */
	public TMfactory() throws Exception
	{
		this( initPS(), -1);
	}
	
	/* Make a factory and force it to use a specific header type.
	 *  Opens a PacketSource object for use in communication
	 *   with the external TOStype device.
	 * 
	 * @param t  -- Force specific type, mostly for extended headers,
	 * 				 Values are defined in TOS_Msg, or use -1 for none.
	 */
	public TMfactory( int t ) throws Exception
	{
		this( initPS(), t );
	}
	
	/** Create a factory for our TOSMSG header type.
	 *   Use the t argument to force a specific header type,
	 *   or use the ps argument to get the MOTECOM type.
	 * 
	 * @param ps -- Packet source will contain the MOTECOM type
	 * @param t  -- Force specific type, mostly for extended headers,
	 * 				 Values are defined in TOS_Msg, or use -1 for none.
	 */
	public TMfactory( PacketSource ps, int t ) throws Exception
	{
		super();
		
		packetSource = ps;
		
		// get the type we want
		if( t == -1 )
			headerType = ps.getPlatform();
		else
			headerType = t;
		
		// setup logging mechanism
		makeLog();
		
		// crank up the message reader
		startReadThread();
	}
	
	/** Open the TOS PacketSource object using the MOTECOM envvar.
	 *   Needs to be static so constructors can call it.
	 * @return PacketSource or else blow cookies
	 */
	static protected PacketSource initPS() throws Exception
	{
		// get a TOSish connection
		PacketSource ps = BuildSource.makePacketSource();
		if( ps == null )
		{
			if( errmsg )
				System.err.println(
	"Invalid packet source (check your MOTECOM environment variable)");
			
			throw new IOException();
		}
		
		// this may also toss IOException cookies...
		ps.open(PrintStreamMessenger.err);
		
		// set decent read ack timeout value
		//  note: requires final->public mod in Packetizer.java
//		((Packetizer)ps).ACK_TIMEOUT = 10;	// for micaz
		
		return ps;
	}

	// stub for creating messages of our type
	// need this to get to header parsing code
	//  to find type for a received buffer
	protected TOS_Msg msgMaster = null;
	
	/** Create a Message class of the right kind for the given type.
	 *   Switch on the message AM types to be expected.
	 *   This default impl just makes stubs for the underlying struct.
	 *   Application should extend this to make specific message types
	 *    and call this only for unknown types and default settings.
	 *    
	 *   @param AMtype -- user defined message type, -1 for default
	 *   @return -- a new message of the right type or the default STUB
	 */
	public TOS_Msg makeMessage( int AMtype )
	{
		// make a default stub type
		// based on HDRtype that we got from the PackSource
		switch( headerType )
		{
			case TMheader.TMCC1000:
			case TMheader.TMEYES:	// ?? dunno
				return new TMstub1000();
				
			case TMheader.TMCC2420t:	// telos
			case TMheader.TMCC2420z:	// micaz
				return new TMstub2420();
				
			case TMheader.TMCC1000XM:
				return new TMstubXM1000();
				
			case TMheader.TMCC2420XM: // not impl yet...
				return new TMstub2420();
				
			case TMheader.TMUPIC:
				return new TMstubUPIC();
			
			default:
				return null;
		}
	}
    
	/** Do something with the current received message 
	 *   ...extend to do something useful...
	 *   ...maybe put in message queue or something...
	 */
    synchronized public void doMessage( TOS_Msg msg )
    {
		curMsg = msg;
		return;
    }
    
	/** create the logging mechanism to use for send and receive
	 *   ...extend to do other stuff...
	 */
    public void makeLog()
    {
		// set a message logging mechanism 
		MessageLog log = new MessageLogOne( false, System.out );
		this.setLogging( log );
		return;
    }
	
	/** Create a Message class of the right kind for the given type.
	 *  This presumes that the readThread is running because the
	 *   msgMaster field is instantiated there...
	 *  Uses the above method as extended by user to make right type.
	 *    
	 *   @param packet -- full received data packet
	 *   @return -- a new message of the right type or the default STUB
	 */
	public TOS_Msg makeMessage( byte[] packet )
	{
		TOS_Msg msg = makeMessage( msgMaster.getMsgType( packet ) );
		
		if( msg != null )
		{
			// install and parse the packet in the message
			msg.setDataBuffer( packet );
		}
		return msg;
	}
	
	protected boolean runme = false;	// in order to kill thread cleanly
	protected Thread threadID;			// thread ID for housekeeping
	protected TOS_Msg curMsg = null;	// current read message
	protected Integer lock = new Integer(0); // sync access to log
	protected MessageLog log = null;	// logging scheme to use

	/** Make a new thread to read from Mote
	 *   May call setLogging() to set some kind of log mechanism later.
	 */
	protected void startReadThread()
	{
		runme = true;	// set to false to exit thread
		msgMaster = makeMessage(-1); // get default stub
		
        // start the read thread
        threadID = new Thread( this, "ReadThread" );
        threadID.start();
	}
	
	/** set the logging mechanism to use when messages are received
	 *      sync may have problems with write/read
	 *      access when the message timing is really short.
	 * 
	 * @param l -- new logging object
	 */
	public void setLogging( MessageLog l )
	{
		synchronized( lock )
		{
			log = l;
		}
	}
	/** turn on/off lowlevel tracing 
	 * 
	 * @param en -- true enables tracemsg and errmsg
	 */
	public void setTrace( boolean en )
	{
		tracemsg = en;
		errmsg = en;
	}
	
	/** Return the current message structure or null if none.
	 *   ...could extend to get message from queue or something...
	 */
    synchronized public TOS_Msg getMessage()
    {
    	return curMsg;
    }

    /** terminate with extreme prejudice
     *   force run() to break as best we can
     */
    public void killReadThread()
    {
        runme = false;

        // neither of interrupt or stop seem to work on NT....
        // a blocked Stdin.read() just never breaks w/o calling exit.
        // but we try them anyway...
        threadID.interrupt();  // does this work?..apparently....
    }

	/** run the read thread
	 *   read mote messages,
	 *   create correct tyep of TOS_Msg message
	 *   log results using logger specified above.
	 *   call doMessage() so user can do something
	 *  bails out on errors or when runme is falsified
	 */
    public void run()
    {
		byte[] packet = null;
    	TOS_Msg msg;
		try
		{
			while( runme &&
				  ((packet = packetSource.readPacket()) != null) )
			{
				// get the right message type wrapper
				msg = makeMessage( packet );
				
				if( tracemsg )
					System.out.println( "RX: " + msg.toHexString() );
				
				// do the sync thing incase we try to replace
				//  the log mechanism when running...
				synchronized( lock )
				{
					// log message if we have a mechanism
					if( log != null )
						log.logRead( msg );
				}
				
				// do some user thing with the message
				doMessage( msg );
			}

		} catch( Exception e )
		{
			if( tracemsg )
				System.err.println( "ReadThread exception on " +
						packetSource.getName() + " -- " +
						e.toString() + " : " + e.getMessage() );
		}

		if( errmsg )
			System.err.println(
				"Dude: The Message Read Thread has bailed out." );
		
		return;
	}
    
    public boolean writeMessage( TOS_Msg cmdMsg )
    {
		boolean rval = false;
		
		// get data buffer to send -- also may set contents
		byte[] data = cmdMsg.getDataBuffer();
		
		// dump contents of buffer
		if( tracemsg )
			System.out.println( "TX: " + cmdMsg.toHexString() );
		
		// log message if so desired
		if( log != null)
			log.logWrite( cmdMsg );
		
		try
		{
			rval = packetSource.writePacket( data );
			
			if( (rval == false) && errmsg )
				System.err.println( "Message write failed." );
			
		} catch( IOException e )
		{
			if( errmsg )
				System.err.println( "Message write failed: " +
						e.toString() + " : " + e.getMessage() );
		}
		
		return rval;
    }
	
} // end'o'TMFactory
