/** TMheader.java -- Base class for TOS Message header fields.
 * 		Utilities to parse and set specific "hardware" message types.
**/
package com.etantdonnes.tinyos;

public abstract class TMheader
{
	// possible TOS message header types -- where relevant
	//  the values match those in net.tinyos.packet.Platform
	public static final int TMCC1000 = 1;	// mica2 radio and serial
	public static final int TMCC2420t = 2;	// telsob radio
	public static final int TMCC2420z = 3;	// micaz radio (same as telos)
	public static final int TMEYES = 4;		// PLATFORM_EYES (??)
	public static final int TMCC1000XM = 50;// Xmesh with mica2/serial
	public static final int TMCC2420XM = 51;// Xmesh with micaz/telsob radio
	public static final int TMUPIC = 100;	// schip's USBPIC 
	
	// default values
	public static final int BCASTADDR = 0xffff;	// broadcast destAddr
	public static final int GROUPID  = 0x7d;	// default group in Makelocal

	// Common components of all Mote messages
	//   (contents of mica2 and serial packet header in TOS1)
	// message AM type -- specific to each child message TYPE
	protected int handlerID = -1;
	// actual size of payload minus header (parsed user struct length)
	//  may be the same as Paysize
	protected int msgLen = -1;
	
	// TODO: legacy from CC10000 TOS_Msg, maybe move there?
	// destination MoteID
	protected int destAddr = -1;
	// mote group ID (assigned in TOS make)
	protected int groupID = -1;
	
	
	/** Create header and leave base values unset
	 */
	protected TMheader() {  super();  }
	
	/** Create header and set base values to given:
	 * @param type	-- message AM type
	 * @param len	-- actual message payload length
	 * @param dest	-- destAddr
	 * @param gid	-- groupID
	 */
	protected TMheader( int type, int len, int dest, int gid )
	{
		this();
		this.handlerID = type;
		this.msgLen    = len;
		this.destAddr  = dest;
		this.groupID   = gid;
	}
	
	/** Allocate a full message buffer for this kind of message.
	 *  Note: this is the maximum buffer size:
	 *    header fields + message payload
	 *    e.g., header fields + msgLen or TOSH_DATA_LENGTH
	 *    
	 * @return -- buffer
	 */
	protected byte[] allocBuf()
	{
		return new byte[getMsgLen()+ getPaystart()];
	}

	/** Convert header values to pretty print format.
	 **  override if needed...
	public String toPrettyString()
	{
		// @todo  -- punt for now...
		return toString();
	}
	 **/

	/** Convert header data buffer to hex-byte String for raw printing.
	 *   Append to given StringBuffer.
	 * @param buf -- StringBuffer to use
	 * @param b -- buffer to read
	 * @param i -- starting byte
	 * 
	 * @return -- String
	 */
	public String toHexString( StringBuffer buf, byte[] b, int i )
	{
		int e = getPaystart();
		while( i < e )
			buf.append( parseHelp.b2Hex( b[i++] ) + " " );
		
		return buf.toString();
	}

	/** Convert header values to readable format.
	 **/
	public String toString()
	{
		return( "h=" + handlerID + " " +
				"l=" + msgLen + " " +
				"d=" + destAddr + " " +
				"g=" + groupID );
	}
	
	/** Return the message header type as pre-defined
	 * @return -- class type
	 */
	public abstract int getHeadtype(); // {  return Headtype;  }
	
	/** Return the starting byte of the user payload message area.
	 **/
	public abstract int getPaystart(); // {  return Paystart;  }
	
	/** Return the length of the user payload message area.
	 **/
	public abstract int getPaySize(); // {  return Paysize;  }

	/** Parse local header fields from given data buffer.
	 *  Used to get header values from a data buffer, on receive.
	 *  @param b -- byte array containing data read from device
	 *  @param s -- parse starting offset in buffer
	 * @return -- number of bytes parsed (next byte position in buffer)
	 * 			  or -1 on errors
	 */
	protected abstract int parseHeader( byte[] b, int s );
	
	/** Set local header fields into given data buffer.
	 *  Used to initialize header values in data buffer, on send.
	 *  @param b -- byte array to be set
	 *  @param s -- starting offset in buffer
	 * @return -- number of bytes set (next byte position in buffer)
	 * 			  or -1 on errors
	 **/
	protected abstract int setHeader( byte[] b, int s );
	
	/** Return the message AM type from the given data packet
	 *   so we know what kind of payload message to use.
	 *  Needs to know where the handlerID is in the header buffer...
	 * @param p -- raw message packet
	 * @return -- handlerID/type field
	 */
	protected abstract int getHandlerID( byte[] p );
	
/********************** basic accessors *****************************/
	/** @return Returns the handlerID field. */
	public int getHandlerID() { return handlerID; }
	
	/** @param handlerID -- The handlerID to set. */
	protected void setHandlerID( int handlerID )
	{ this.handlerID = handlerID; }
	
	/** Return the message length from the given data packet
	 *   if msgLen is not set then return the default payload size.
	 *   This may be the actual user message struct length and
	 *    not the header msgLen, c.f. XMesh overlay...
	 * @return -- user struct length
	 */
	public int getMsgLen()
	{
		// if the header has been parsed then msgLen will be set,
		// otherwise use the defined max payload size
		if( this.msgLen != -1 )
			return this.msgLen;
		else
			return this.getPaySize();
	}

	/** @param msgLen -- The message length to set. */
	protected void setMsgLen( int msgLen )
	{ this.msgLen = msgLen;	}

	/** @return Returns the destAddr field. */
	public int getDestAddr() { return this.destAddr; }

	/** @param destAddr -- The destAddr to set. */
	protected void setDestAddr( int destAddr )
	{ this.destAddr = destAddr; }

	/** @return Returns the groupID field. */
	public int getGroupID() { return this.groupID; }

	/** @param groupID -- The groupID to set. */
	protected void setGroupID( int groupID )
	{ this.groupID = groupID; }
	
} //end'o'TMheader