/** TMheader1000.java -- Header class for TOS1 CC1000/serial message structures.
 * 		Parse and set specific "hardware" message fields.
 *		  mica2 native in tos/types/AM.h ::
 *		      uint16_t addr; -- destination
 *			  uint8_t type; -- "handlerID" message type 
 *			  uint8_t group; -- syustem groupID, nominal value
 *			  uint8_t length; -- actual payload length
 *			  int8_t data[TOSH_DATA_LENGTH]; -- full payload buffer
 *			  uint16_t crc; -- ignored here...
 *		TOSH_DATA_LENGTH = 29 bytes = PAYSIZE
 *	
 * Rewanked April 2012
**/
package com.etantdonnes.tinyos;

public class TMheader1000 extends TMheader
{
	// sizes and defines
	public static final int MSGSIZE = 34;	// max bytes in entire message
	public static final int PAYSTART =  5;	// start of payload bytes
	public static final int PAYSIZE = 29;	// max payload bytes
	public static final int HEADTYPE = TMheader.TMCC1000; // type
	
	protected static final int USERLEN = 4;	// position of msg length field
	protected static final int USERTYPE = 2;	// position of msg type field

	
	/** Create a thingie for the TMCC1000 header type.
	 *   used to make receive message wrappers
	 **/ 
	public TMheader1000()
	{
		super();
		setMsgLen( 0 );		// just the header...can override later
	}
	
	/** Create a TMCC1000 thingie and set the basic fields.
	 *    used to make send message classes
	 *
 	 *   default destAddr to Broadcast because we can't have two
	 *    constructors with 4 int args each...
	 * 
	 * @param type	-- message AM type
	 * @param len	-- message payload length
	 * @param dest  -- destination mote ID or -1 for broadcast
	 * @param gid	-- groupID
	 */
	public TMheader1000( int type, int len, int dest, int gid )
	{
		super( type, len, dest, gid );
	}

	/** parse common fields from given data buffer
	 *  @param b -- byte array containing data read from Mote
	 *  @param s -- parse starting offset in buffer
	 * @return -- next byte to parse or -1 on error
	 */
	public int parseHeader( byte[] b, int s )
	{
		int rval = -1; // Fail
		
		// check the buffer and message lengths just in case
		//  payload length is at byte 4 for CC1000
		if( (b.length <= MSGSIZE) &&		// buffer size
			((b[s+USERLEN] & 0xff) <= PAYSIZE) )	// max payload length
		{
			// get 'standard' Mica2 message header
			this.destAddr  = parseHelp.ti(b, s); s+=2;
			this.handlerID = parseHelp.tb(b, s); s+=1;
			this.groupID   = parseHelp.tb(b, s); s+=1;
			this.msgLen    = parseHelp.tb(b, s); s+=1;

			rval = s;	// A-OK
		}

		return rval;
	}
	
	/** Set common fields in given data buffer
	 *  @param b -- byte array to set
	 *  @param s -- starting offset in buffer
	 * @return -- next byte to set or -1 on error
	 **/
	public int setHeader( byte[] b, int s )
	{
		// set 'standard' Mica2 message header
		parseHelp.fi( this.destAddr, b, s );  s+=2;
		parseHelp.fb( this.handlerID, b, s ); s+=1;
		parseHelp.fb( this.groupID, b, s );   s+=1;
		parseHelp.fb( this.msgLen, b, s );    s+=1;

		return s;
	}
	
	/** Return the message AM type from the given data buffer
	 *   so we know what kind of message to use.
	 * @param p -- raw message packet
	 * @return -- type field
	 */
	public int getHandlerID( byte[] b )
	{
		return b[USERTYPE] & 0xff;
	}
	
	/** Convert header data buffer to hex-byte String for raw printing.
	 *   Append to given StringBuffer.
	 * @param buf -- StringBuffer to fill
	 * @param b -- buffer to read
	 * @param s -- starting byte
	 * @return -- StringBuffer
	 * (non-Javadoc)
	 * @see com.etantdonnes.tinyos.TMheader#toHexString()
	public String toHexString( StringBuffer buf, byte[] b, int s )
	{
		/*
		 out.print( hdr + "(" + p.length + ")" +
		 " d=" + (tm.ti(p,0) & 0xffff) +
		 " t=" +       (p[2] & 0xff) +
		 " g=" +       (p[3] & 0xff) +
		 " l=" +       (p[4] & 0xff) +
		 " ::\t"
		 );
		 */
	/* 
		// first field is LSB shortint...
		buf.append( helpParse.b2Hex( b[s+1] ) );
		buf.append( helpParse.b2Hex( b[s+0] ) + " " );
		// the rest are single bytes
		buf.append( helpParse.b2Hex( b[s+2] ) + " " );
		buf.append( helpParse.b2Hex( b[s+3] ) + " " );
		buf.append( helpParse.b2Hex( b[s+4] ) + " :: " );

		 //for( int i=s+2; i < p.length; ++i )
		 //buf.append( Integer.toHexString( b[i] ) + " " );

		return buf.toString();
	}
	 */
		
	/** Convert local header values to readable format
	 **/
	public String toString()
	{
		return(
				"d="  + this.destAddr +
				" t=" + this.handlerID +
				" g=" + this.groupID +
				" l=" + this.msgLen );
	}

	@Override
	public int getHeadtype() { return HEADTYPE; }
	@Override
	public int getPaySize() { return PAYSIZE; }
	@Override
	public int getPaystart() { return PAYSTART; }
	
} //end'o'TMheader1000