/** TMheader2420.java -- Header class for TOS CC2420 message structures.
 * 		Parse and set specific "hardware" message fields.
 *		 telos and micaz native AM.h ::
 *			uint8_t length;
 *			uint8_t fcfhi;
 *			uint8_t fcflo;
 *			uint8_t dsn;
 *			uint16_t destpan;
 *			uint16_t addr;
 *			uint8_t type;
 *			uint8_t group;
 *			int8_t data[TOSH_DATA_LENGTH];
 *		    uint16_t crc; -- ignored here...
 *		TOSH_DATA_LENGTH = 29 bytes = PAYSIZE
 *
 * Rewanked April 2012
**/
package com.etantdonnes.tinyos;

public class TMheader2420 extends TMheader
{
	// sizes
	public static final int MSGSIZE = 39;	// max bytes in entire message
	public static final int PAYSTART = 10;	// start of payload bytes
	public static final int PAYSIZE = 29;	// max payload bytes in packet
	public static final int HEADTYPE = TMheader.TMCC2420t; // type
	
	protected static final int USERLEN = 0;	// position of msg length field
	protected static final int USERTYPE = 8;	// position of msg type field
	
	// added fields of the standard Tmote/Micaz message
	public int fcfhi;
	public int fcflo;
	public int dsn;
	public int destpan;
	
	/** Create a thingie, for internal use
	 *    All the header type parameters get set here.
	 * 
	 * @param pks -- user payload packet size
	protected TMheader2420( int psz )
	{
		super( TMheader.TMCC2420t, MSGSIZE, PAYSTART, psz, crap );
	}
	 */
	
	/** Create a thingie for the TMCC2420 header type.
	 *   used to make receive message wrappers
	 **/ 
	public TMheader2420()
	{
		super();
		setMsgLen( 0 );		// just the header...can override later
	}
	
	/** Create a TMCC2420 thingie and set the basic fields.
	 * 		used to make send messages
	 * 
	 *   default destAddr to Broadcast
	 * 
	 * @param type	-- message AM type
	 * @param len	-- message payload length
	 * @param dest TODO
	 * @param gid	-- groupID
	 */
	public TMheader2420( int type, int len, int dest, int gid )
	{
		super( type, len, dest, gid );
	}

	/** parse common fields from given data buffer
	 *  @param b -- byte array containing data read from Mote
	 *  @param s -- parse starting offset in buffer
	 * @return -- next byte to parse or -1 on error
	 */
	public int parseHeader( byte[] b, int s )
	{
		int rval = -1; // Fail

		// check the buffer and message lengths just in case
		//  payload length is at byte 0 for 2420
		if( (b.length <= MSGSIZE) &&		// buffer size
			((b[s+USERLEN] & 0xff) <= PAYSIZE) )	// payload length
		{
			// get MicaZ message header
			this.msgLen    = parseHelp.tb(b, s); s+=1;
			this.fcfhi     = parseHelp.tb(b, s); s+=1;
			this.fcflo     = parseHelp.tb(b, s); s+=1;
			this.dsn       = parseHelp.tb(b, s); s+=1;
			this.destpan   = parseHelp.ti(b, s); s+=2;
			this.destAddr  = parseHelp.ti(b, s); s+=2;
			this.handlerID = parseHelp.tb(b, s); s+=1;
			this.groupID   = parseHelp.tb(b, s); s+=1;

			rval = s;	// A-OK
		}

		return rval;
	}
	
	/** Set common fields in given data buffer
	 *  @param b -- byte array to set
	 *  @param s -- starting offset in buffer
	 * @return -- next byte to set or -1 on error
	 **/
	public int setHeader( byte[] b, int s )
	{
		parseHelp.fb( this.msgLen,  b, s );    s+=1;
		parseHelp.fb( this.fcfhi,  b, s );     s+=1;
		parseHelp.fb( this.fcflo,  b, s );     s+=1;
		parseHelp.fb( this.dsn,  b, s );       s+=1;
		parseHelp.fi( this.destpan,  b, s );   s+=2;
		parseHelp.fi( this.destAddr, b, s );   s+=2;
		parseHelp.fb( this.handlerID,  b, s ); s+=1;
		parseHelp.fb( this.groupID,  b, s );   s+=1;
			
		return s;
	}
	
	
	/** Return the message AM type from the given data buffer
	 *   so we know what kind of message to use.
	 * @param p -- raw message packet
	 * @return -- type field
	 */
	public int getHandlerID( byte[] b )
	{
		return b[USERTYPE] & 0xff;
	}
	
	/** Convert header data buffer to hex-byte String for raw printing.
	 *  TODO: can probably use base class...
	 *   Append to given StringBuffer.
	 * @param buf -- StringBuffer to fill
	 * @param b -- buffer to read
	 * @param s -- starting byte
	 * @return -- StringBuffer
	 * (non-Javadoc)
	 * @see com.etantdonnes.tinyos.TMheader#toHexString()
	public String toHexString( StringBuffer buf, byte[] b, int s )
	{
		buf.append( helpParse.b2Hex( b[s+0] ) + " " );
		buf.append( helpParse.b2Hex( b[s+1] ) + " " );
		buf.append( helpParse.b2Hex( b[s+2] ) + " " );
		buf.append( helpParse.b2Hex( b[s+3] ) + " " );
		
		// two byte LSB int
		buf.append( helpParse.b2Hex( b[s+5] ) );
		buf.append( helpParse.b2Hex( b[s+4] ) + " " );
		
		// two byte LSB int
		buf.append( helpParse.b2Hex( b[s+7] ) );
		buf.append( helpParse.b2Hex( b[s+6] ) + " " );
		
		buf.append( helpParse.b2Hex( b[s+8] ) + " " );
		buf.append( helpParse.b2Hex( b[s+9] ) + " :: " );
		
		return buf.toString();
	}
	 */
		
	/** Convert local header values to readable format
	 **/
	public String toString()
	{
		return(
				"l=" + this.msgLen +
	            " i=" + this.fcfhi +
				" o=" + this.fcflo +
				" s=" + this.dsn +
				" p=" + this.destpan +
				" d="  + this.destAddr +
				" t=" + this.handlerID +
				" g=" + this.groupID  );
	}
	
	@Override
	public int getHeadtype() { return HEADTYPE; }
	@Override
	public int getPaySize() { return PAYSIZE; }
	@Override
	public int getPaystart() { return PAYSTART; }

}//end'o'TMheader2420