/** TMheaderUPIC.java -- Header class for schip's USBPIC message structures.
 * 		Parse and set specific "hardware" message fields.
		 upic message structure:
		  -------------------------- added by Packetizer.java framer protocol
		 	uint8_t frame=0x7E;
		 	uint8_t mtype=0x42;	//(66 dec)
		 	uint8_t seqno;
		  -------------------------- user message
			uint8_t length;					// length of user data
			uint8_t type;					// user message type
			int8_t data[MAX_UPICPAY_LEN];	// payload: PAYSIZE
		  -------------------------- added by Packetizer.java framer protocol
		 	uint8_t frame=0x7E;
		  --------------------------
		This class only deals with the "user message" bytes.
**/
package com.etantdonnes.tinyos;

public class TMheaderUPIC extends TMheader
{
	// sizes
	public static final int MSGSIZE = 60;	// max bytes in entire user message
	public static final int PAYSIZE = 58;	// max payload bytes in packet
	
	protected static final int USERLEN = 0;	// position of msg length field
	protected static final int USERTYPE = 1;	// position of msg type field
	public static final int PAYSTART = 2;	// start of payload bytes
	public static final int HEADTYPE = TMheader.TMUPIC; // type
	
	/** Create a thingie for the TMUPIC header type.
	 *  Used to make receive message wrappers.
	 *   Defaults to 0 payload size until 
	 *    message header is parsed or setMsgLen() is called.
	 **/ 
	public TMheaderUPIC()
	{
		super();
		setMsgLen( 0 );		// just the header...can override later
	}
	
	/** Create a TMUPIC thingie and set the basic fields.
	 * 	Used to make send messages.
	 *   Because they are not used in the UPIC, default
	 *    destAddr to BCASTADDR and groupID to GROUPID
	 *   
	 * @param type	-- message AM type
	 * @param len	-- actual message payload length
	 */
	public TMheaderUPIC( int type, int len )
	{
		super( type, len, TMheader.BCASTADDR, TMheader.GROUPID );
	}

	/** parse header fields from given data buffer
	 *  @param b -- byte array containing data read from Mote
	 *  @param s -- parse starting offset in buffer
	 * @return -- next byte to parse or -1 on error
	 */
	public int parseHeader( byte[] b, int s )
	{
		int rval = -1; // Fail

		// check the buffer and message lengths just in case
		//  payload length is at byte 0 for UPIC
		if( (b.length <= MSGSIZE) &&		// buffer size
			((b[s+USERLEN] & 0xff) <= PAYSIZE) )	// payload length
		{
			// get UPIC message header
			this.msgLen    = parseHelp.tb(b, s); s+=1;
			this.handlerID = parseHelp.tb(b, s); s+=1;

			// null the unused
			this.destAddr  = -1;
			this.groupID   = -1;

			rval = s;	// A-OK
		}

		return rval;
	}
	
	/** Set common fields in given data buffer
	 *  @param b -- byte array to set
	 *  @param s -- starting offset in buffer
	 * @return -- next byte to set or -1 on error
	 **/
	public int setHeader( byte[] b, int s )
	{
		parseHelp.fb( this.msgLen, b, s );    s+=1;
		parseHelp.fb( this.handlerID, b, s ); s+=1;
		
		return s;
	}
	
	/** Return the message AM type from the given data buffer
	 *   so we know what kind of message to instantiate.
	 * @param p -- raw message packet
	 * @return -- type field
	 */
	public int getHandlerID( byte[] b )
	{
		return b[USERTYPE] & 0xff;
	}
	
	/** Convert local header values to readable format
	 **/
	public String toString()
	{
		return( "l=" + this.msgLen +
				" t=" + this.handlerID );
	}

	@Override
	public int getHeadtype() { return HEADTYPE; }
	@Override
	public int getPaySize() { return PAYSIZE; }
	@Override
	public int getPaystart() { return PAYSTART; }
	
}//end'o'TMheaderUPIC