/* TMXMheader1000.java --
 *  Header for Xbow Xmesh messages using mica2 type radio.
 *   A concatenation of mica2 and xmesh headers.
		uint16_t  sourceaddr  Address of mote that sent the message 
		uint16_t  originaddr  Address of mote that originated the message 
		uint16_t  seqno       Message sequence number.  
		uint8_t   socket      Application message type ID 
 * 
 * Created on Jun 25, 2009
 */
package com.etantdonnes.tinyos;

public class TMheaderXM1000 extends TMheader1000
{
	// sizes
	public static final int MSGSIZE = 54;	// max bytes in entire message
	public static final int PAYSTART = 12;	// start of payload bytes after XM
	public static final int PAYSIZE = 29;	// max payload bytes
	public static final int HEADTYPE = TMheader.TMCC1000XM; // type
	
	// extended user type field in XMesh message
	protected static final int USERTYPE = 11;	// position of "socket" field
	protected static final int XMLEN = 7;		// length of added XMesh header
	
	/** XMesh header components **/
	public int sourceAddr;
	public int originAddr;
	public int seqno;
	public int socket;
	
	/** Create a thingie for the TMCC2420 XMesh header type.
	 *   used to make receive message wrappers
	 **/
	public TMheaderXM1000()
	{
		super();
		setMsgLen( 0 );		// just the header...can override later
	}
	
	/** Create a TMCC1000 Xmesh thingie and set the basic fields.
	 * 		used to make send messages
	 * 
	 * TODO: probably need to massage type setting correctly
	 * 			vis the socket thing, and set our local XMesh values
	 *
	 * @param type	-- message AM type
	 * @param len	-- message payload length
	 * @param dest  -- destination mote address
	 * @param gid	-- groupID
	 */
	public TMheaderXM1000( int type, int len, int dest, int gid )
	{
		super( type, len, dest, gid );
	}
	
	/** Return the user defined message type from the given packet.
	 * @param p -- raw message packet
	 * @return -- type field
	 * 
	 * XMesh system messages will be their AM type.
	 * User data messages will be their "socket" type.
	 *  To simplify the issue,
	 *    all user message types are in these ranges:
	 *    	5-10, 20-80  -- see TMMXmsgtype for XMesh types.
	 *  
	 */
	public int getHandlerID( byte[] b )
	{
		// if this is an XMesh housekeeping message
		// just return the base AM type
		// otherwise we dig deeper to the user message type
		int amt = super.getHandlerID( b );
		if( (amt == TMMXmsgtype.AM_DATA2BASE) ||
		    (amt == TMMXmsgtype.AM_DATA2NODE) ||
		    (amt == TMMXmsgtype.AM_DATAACK2BASE) ||
		    (amt == TMMXmsgtype.AM_DATAACK2NODE) )
		{
			amt = b[USERTYPE] & 0xff;
		}
		
		return amt;
	}

	/** Return the message length from the given data packet
	 *   this is the actual user message struct length and
	 *   is not the header msgLen due to the XMesh overlay...
	 * @return -- user struct length
	 */
	public int getMsgLen()
	{
		int len = super.getMsgLen();
		return len - XMLEN;
	}
	
	/** parse common fields from given data buffer
	 *  @param b -- byte array containing data read from Mote
	 *  @param s -- parse starting offset in buffer
	 * @return -- next byte to parse or -1 on error
	 */
	public int parseHeader( byte[] b, int s )
	{
		// parse micaz header and get the first byte of our header area
		s = super.parseHeader( b, s );
		if( s < 0 )		return s;
		
		// get XMESH message header (7 additional bytes)
		this.sourceAddr   = parseHelp.ti(b,s+0); s+=2;
		this.originAddr   = parseHelp.ti(b,s+2); s+=2;
		this.seqno        = parseHelp.ti(b,s+4); s+=2;
		this.socket       = parseHelp.tb(b,s+6); s+=1;
		
		return s;
	}
	
	/** Set common fields in given data buffer
	 *  @param b -- byte array to set
	 *  @param s -- starting offset in buffer
	 * @return -- next byte to set or -1 on error
	 * 
	 * Note that we only set the socket value
	 *  as the others are for internal system management.
	 **/
	public int setHeader( byte[] b, int s )
	{
		// set header and get the first byte of the payload area
		s = super.setHeader( b, s );
		if( s < 0 )		return s;
		
		// set XMESH message header (7 additional bytes)
		//this.sourceAddr   = helpParse.ti(b,s+0) & 0xffff;
		//this.originAddr   = helpParse.ti(b,s+2) & 0xffff;
		//this.seqno        = helpParse.ti(b,s+4) & 0xffff;
		// b[s+6] = (byte)(this.socket & 0xff);
		parseHelp.fb( this.socket, b, s+6 );
			
		return s+7;
	}
	
	/** Convert local header values to readable format
	 **/
	public String toString()
	{
		return( super.toString() + 
				"s=" + this.sourceAddr +
	            " o=" + this.originAddr +
				" n=" + this.seqno +
				" a=" + this.socket );
	}
	
	/** Set the user defined message type for XMesh messages.
	 *   TODO: need to re-wank this...
	 * @param t -- User message sub-AM type
	 */
	public void setUserMsgType( int t )
	{
		this.socket = t;
	}
	
	@Override
	public int getHeadtype() { return HEADTYPE; }
	@Override
	public int getPaySize() { return PAYSIZE; }
	@Override
	public int getPaystart() { return PAYSTART; }

} //end'o'TMheaderXM1000
