/** TOSComm.java -- Driver program for TMote and TOSComm communications.
 *    Doc TBD...
 *   
 *   Version 2.0
**/
package com.etantdonnes.tinyos;

import java.io.*;


@SuppressWarnings("unchecked") // get rid of TOS_Msg check whines
public class TOSComm
{
	/** main execution
	 *   creates a Message Factory and runs TOSComm
	 *   override to make your own
	 * 
	 * @param args
	 * @throws Exception
	 */
	public static void main(String args[]) throws Exception
	{
		try
		{
			// get my xxComm object
			//  sets TOSComm "global" String programName
			TOSComm tc = new TOSComm();
			
			// parse args and get an input stream
			BufferedReader in = tc.parseArgs( args );
			
			// get my TOS connection and Message Factory
			//  uses MOTECOM environment variable to find port and type
			TMfactory mf = new TMfactory(); // recv msg factory
			
			// execute input function interpreter
			System.out.println( "starting " + programName );
			tc.execute( mf, in );
			System.out.println( programName + " done" );
		
		} catch( Exception e )
		{
			System.err.println( programName + " Error initializing" +
								" -- " + e.toString() +
								" : " + e.getMessage() );
		}	
		
		System.exit(0);
	} // end'o'main
		

	// Note: these can be [re]set in setUp()...
	// master program name
	static protected String programName = "TOSComm";
	// get/save file name
	static protected String DefaultFile = "TCsave.txt";
	// default values for TOS...
	static protected int destAddr = TMheader.BCASTADDR;
	static protected int groupID  = TMheader.GROUPID;
	
	/** create the whole mgilla
	 */
	public TOSComm()
	{
		super();
		setUp();
	}
		
	/** parse command line arguments and return an input Reader
	 *   The only arg we expect here is a command input file name,
	 *    otherwise we return a BufferedReader on System.in 
	 * @param args
	 * @return -- command input stream
	 * @exception -- file open errors
	 */
	public BufferedReader parseArgs( String args[] )
									throws Exception
	{
		BufferedReader cmdReader = null;
		// look at command args and open input source
		if( args.length > 1 )
		{
			System.err.println(
				"usage: java " + programName + " [infile]" );
			System.exit(2);
		}
		else if (args.length == 1)
		{
			// open an optional command input file
			cmdReader =  new BufferedReader(
					new FileReader( args[0] ));
		}
		else
		{
			// default to getting input from command line
			cmdReader = new BufferedReader(
					new InputStreamReader( System.in ) );
		}
		
		return cmdReader;
	}
	
	/** master program execution 
	 ** Console driven test program for TOS_Msg communications.
	 *  Returns when input stream is closed, or on quit command.
	 * 
	 * @param msgFactory -- message source/sink
	 * @param instream -- BufferedReader input source,
	 * 					   nominally a massaged System.in
	 */
	public void execute( TMfactory msgFactory, BufferedReader instream )
	{
		try
		{
			// command and status messages
			// get stub messages of our type for sending
			TOS_Msg statMsg = makeStatusMessage( msgFactory );
			TOS_Msg cmdMsg = makeCommandMessage( msgFactory, null );
			TOS_Msg sendMsg = null;	// the actual message to send
			
			StreamTokenizer inp = null;
			StreamTokenizer save_inp = null;
			BufferedReader filereader = null;
			
			// command param parsing array,
			//  this may be 2x as large as needed because
			//  NUMFIELDS is probably in bytes for the stubs
			int cparams[] = new int[cmdMsg.getNumFields()];
			
			// Note: STREAM-f*****g-Tokenizer isn't able 
			//	to set delims other than whitespace right...
			//  so it's all WS and no commas from here.
			//  No wonder it's deprecated anyway...
			inp = new StreamTokenizer( instream );
			
			// set a bunch of options to make things work nice
			inp.eolIsSignificant( true );
			inp.lowerCaseMode( true );
			inp.slashSlashComments( false );
			inp.slashStarComments( false );
			inp.parseNumbers();
	
			boolean run = true;		// run the input loop while true
			
			while( run )
			{
				// only execute when we have a message to send
				sendMsg = null;
				
				// print a prompt
				System.out.print( "?>> " );
	
				// get something from the inputstream
				int ttype = inp.nextToken();
				// if it is text, then get the first char as a command
				if( ttype == StreamTokenizer.TT_WORD )
					ttype = (int) inp.sval.charAt( 0 );
	
				// get the parameter values to End Of LINE
				//  set cparams to either a decimal number or
				//  the first char of each space delimited field
				int i = 0;  // parameter counter
				int tt = ttype;
				while(  (tt != StreamTokenizer.TT_EOL) &&
						(tt != StreamTokenizer.TT_EOF) &&
						(i < cparams.length) )
				{
					tt = inp.nextToken();
					if( tt == StreamTokenizer.TT_NUMBER )
						cparams[i++] = (int) inp.nval;
					else if( ttype == StreamTokenizer.TT_WORD )
						cparams[i++] = (int) inp.sval.charAt(0);
				}
				
				// null unused param values on each pass
				while( i < cparams.length )
				{
					cparams[i++] = -1;
				}
				
				// do something with it
				switch( ttype )
				{
					// print command help
					case '?':
					{
						System.out.println( extendHelp() );
					}
					break;
	
					// bailout when we quit or on input end
					case 'q':
					case StreamTokenizer.TT_EOF:
					{
						// if we're at the end of a file
						//  swap the input stream back in
						// otherwise bail out WExP
						if( save_inp != null )
						{
							inp = save_inp;
							filereader.close();
							filereader = null;
							save_inp = null;
						}
						else
						{
							//System.out.println( "bye bye" );
							run = false;
						}
					}
					break;
	
					case 'd':	// set destination and groupID
					{
						if( cparams[0] != -1 )
							destAddr = cparams[0];
						if( cparams[1] != -1 )
							groupID = cparams[1];
	
						statMsg.setDestAddr( destAddr );
						statMsg.setGroupID( groupID );
						cmdMsg.setDestAddr( destAddr );
						cmdMsg.setGroupID( groupID );
						
						System.out.println( "destAddr: " + destAddr +
										    " groupID: "  + groupID );
					}
					break;
	
					case 'm':	// get a new type of command message
								//   and set basic parameters:
								//  type length destination group
					{
						cmdMsg = makeCommandMessage( msgFactory,
													 cparams );
						
						// set the local defaults
						cmdMsg.setDestAddr( destAddr );
						cmdMsg.setGroupID( groupID );
	
						System.out.println( "new Message: " +
										cmdMsg.getClass() + "\n" +
										cmdMsg.toPrettyString() );
					}
					break;
	
					case 'p':	// print current received message
					{
						TOS_Msg m = msgFactory.getMessage();
						if( (m != null) )
						{
							System.out.println( "received Message: " +
											m.getClass() + "\n" +
											m.toPrettyString() );
						}
						else
						{
							System.err.println( "no current Message" );
						}
							
					}
					break;
	
					// set params and send command
					case 's':
					{
						cmdMsg.nullFields();
						sendMessageHook( cmdMsg, cparams );
						sendMsg = cmdMsg; // yes, now exec the message
					}
					break;
					
					// set trace messages in send/receive factory
					case 't':
					{
						msgFactory.setTrace( (cparams[0] == 0) ?
												false : true );
					}
					break;
					
					// store last received message in file
					case 'f':
					{
						String filename = DefaultFile;
						
						// if there's an input word it's the filename
						if( (ttype = inp.nextToken()) ==
								StreamTokenizer.TT_WORD )
						{
							filename = inp.sval;
						}
							
						TOS_Msg msg = msgFactory.getMessage();
						if( (msg != null) )
						{
							try
							{
								// get the file's gdm output stream
								PrintStream out =
								  new PrintStream(
								   new BufferedOutputStream(
								    new FileOutputStream(
								     new File( filename ) )), false );
	
								out.println( msg.toString() );
							
								System.out.println(
										"Saved message in: " +
										filename );
								
								out.flush();
								out.close();
								out = null;
							}
							catch( Exception e )
							{
								System.err.println(
										"Save Error: " + e );
							}
						}
						else
						{
							System.err.println( "No current Message" );
						}
					}
					break;
	
					// get commands from file
					case 'g':
					{
						String filename = DefaultFile;
						
						// if it's a word then get parameter value
						if( (ttype = inp.nextToken()) ==
								StreamTokenizer.TT_WORD )
						{
							filename = inp.sval;
						}
						
						try
						{
							// save the old input stream
							// and swap the file for the duration
							filereader = new BufferedReader( 
										new FileReader( filename ) );
							save_inp = inp;
							inp = new StreamTokenizer( filereader );
							
							System.out.println( "Getting commands from: " +
									filename );
						}
						catch( Exception e )
						{
							System.err.println( "File Error: " + e );
						}
					}
					break;
	
					// empty input line send statreq
					case StreamTokenizer.TT_EOL:
						statMsg.nullFields();
						sendMsg = statMsg;
					break;
					
					default:
						// see if the user has over extended
						sendMsg = extendMenu( ttype, cparams );
						if( sendMsg != null )
						{
							sendMsg.setDestAddr( destAddr );
							sendMsg.setGroupID( groupID );
						}
				}
				
				// if we have a command, send it
				if( sendMsg != null )
				{
					msgFactory.writeMessage( sendMsg );
				}
	
			} // while( run );
		} catch( Exception e )
		{
			System.err.println( programName + " Command Loop Error " +
								" -- " + e.toString() +
								" : " + e.getMessage() );
		}	

		// clean up...
		msgFactory.killReadThread();
		
		return;
	}
	
/***** Possible override methods to change function execution *****/
	
	/** do user setup before anything executes
	 *   like changing default names and stuff
	 */
	protected void setUp()
	{
		// our program name
		programName = "TOSComm";
		// get/save file name
		DefaultFile = "TCsave.txt";
		// TOS message destination ID
		destAddr = TMheader.BCASTADDR;
		// TOS message group ID
		groupID  = TMheader.GROUPID;
		
		return;
	}
	
	/** Make the default status message type to send
	 *   when doing "<CR>" function
	 * @param mf -- message factory to use
	 * @return   -- message structure
	 */
	protected TOS_Msg makeStatusMessage( TMfactory mf )
	{
		// just get a STUB of our kind
		return mf.makeMessage( -1 );
	}
	
	/** Make command message to send when doing "send" function
	 *  Called from initialization and "Message" function.
	 *   default is just set params in header fields
	 *   override as needed...
	 *   
	 * @param mf -- message factory to use
	 * @param cp -- command parameters:
	 * 				 if null we make a STUB message type -1
	 *               otherwise cp[0] = type, cp[1] = length
	 * @return   -- message structure
	 */
	protected TOS_Msg makeCommandMessage( TMfactory mf, int[] cp )
	{
		int type = -1;
		int len = -1;
		
		if( cp != null )
		{
			type = cp[0];
			len  = cp[1];
		}
		
		// get one of our type and set length is not -1
		TOS_Msg msg = mf.makeMessage( type );
		if( type != -1 ) // in case we don't have the right factory
			msg.setMsgType( type );
		if( len != -1 )  // in case we use variable length messages
			msg.setMsgLen( len );
		
		return msg;
	}
						
	/** Do something with the command message before sending
	 *  Called from "Send" function.
	 *   default is just set params in message fields
	 *   override as needed...
	 * 
	 * @param cmdMsg -- message to be sent
	 * @param cparams -- follow-on argument values 
	 * @return -- nada
	 */
	protected void sendMessageHook( TOS_Msg cmdMsg, int[] cparams )
	{
		cmdMsg.setFields( cparams );
		return;
	}

	/** Extend the execution menu as needed.
	 *  Called from function switch default:
	 *   when ttype is nor recognized
	 * 
	 * @param ttype	-- user command type (first char of line)
	 * @param cparams -- follow-on argument values 
	 * @return -- TOS_Msg to execute or null if none.
	 */
	protected TOS_Msg extendMenu( int ttype, int[] cparams )
	{
		return null;
	}

	/** Extend the help menu as needed.
	 *   Called from "?" function 
	 *   First call super.extendHelp() then add your stuff...
	 */
	protected String extendHelp()
	{
		return(
		 "Use SPACE delimiters!!\n" +
		 "? -- print this help\n" +
		 "Quit\n" +
		 "<ENTER> -- send empty status request message\n" +
		 "Dest dest group -- set send message params\n" +
		 "Message type len dest group -- set send message params\n" +
		 "Send ... -- set/send command with given values\n" +
		 "Print -- print current received message\n" +
		 "Trace 0/1 -- set low level message debug/trace\n" +
		 "Get filename -- read commands from file\n" +
		 "File filename -- save settings in file\n" );
	}
	
} // end'o'TOSComm