/** TOS_Msg.java -- Basic TOS message structure.
 *		Extend to add specific AM message type behaviors.
 *		Declare TMheader type <T> in final class.
**/
package com.etantdonnes.tinyos;

/** Base class for TOS_Msg communications
 *   abstract because the TOSMsgFactory( int AMtype )
 *    method should be impl...
**/
public abstract class TOS_Msg<T>
{
/** returned by getNumfields() in extended classes...
	private static final int NUMFIELDS = 0;	// parameters in message
**/
	// full message packet buffer (alloc or set in sub-classes)
	protected byte[] data = null;
	
	// header class to parse low-level "hardware" message structures
	//  note: just the header, NOT the user payload
	protected T header = null;

	// helper class for default parsing and printing
	//  behavior controlled by user settings in
	//    getFieldFormat() and getPrintBase()
	protected pHelp fieldFmt;
			
	/** Create an empty thingie.
	 *   Sub-classes may override to
	 *    create and set their preferred header type
	 *    but better to use TOS_Msg(TMheader) below...
	 **/ 
	protected TOS_Msg()
	{
		super();
		fieldFmt =
			parseHelp.getFormatter( getFieldFormat(), getPrintBase() );
	}
	
	/** Create an empty thingie with given header type.
	 *   Sub-classes must create their preferred header type
	 *   and call super(h)...
	 **/ 
	@SuppressWarnings("unchecked")
	protected TOS_Msg( TMheader h )
	{
		this();
		header = (T)h;
	}
	
	
 /**** NOTE: Message Parsing and Setting methods.
  ****   All the offsets and sizes are hard coded here in
  ****  {parse,set}Message() or in {get,set}Fields()
  ****   depending on how one wants to implement ones thingies.
  **** Override to add specific TYPE message handling if desired.
  **** There are three possible use patterns:
  **** 	1. **default** Use data buffer as storage --
  **** 	      parseMessage() just converts header fields from data.
  **** 	      getFields() converts from data into return vals.
  **** 	      setMessage() just converts header fields into data.
  **** 	      setFields() converts from vals into data buffer.
  **** 	2. Use local values array --
  **** 	      parseMessage() converts header and data into int array
  **** 	      getFields() just returns the local int array
  **** 	      setMessage() converts header and int array into data.
  **** 	      setFields() copies given vals into local int array
  **** 	3. Use local member fields --
  **** 	      parseMessage() converts header and data into locals
  **** 	      getFields() copies locals into int array
  **** 	      setMessage() converts header and fields into data
  **** 	      setFields() copies given vals into local values
  ****/

 /**** Utilities to override in sub-classes to change the
  ****   parsing and printing behavior, instead of overriding
  ****   methods for simple format changes.
  ****/
	
	/** Set the format of data buffer interpretation.
	 *   So we don't have to override the actual methods to
	 *    switch between simple data array formats...
	 *   Default to unsigned bytes...
	 *  @return -- any of the helpParse.FORMATS below.
	 */
	protected int getFieldFormat() { return parseHelp.UNSIGNEDBYTE; }
	
	/** Set the format of printed lines in toString().
	 *  Also uses getFieldFormat() to find size of each element
	 *   and [un]signed format.
	 *  Default to eight hex bytes per line...
	 */
	/**  @return -- number of elements to print on a line */
	protected int getFieldsPerLine() { return 8; }

	/**  @return -- print conversion base: 10 or 16 */
	protected int getPrintBase() { return 16; }

	/**  @return -- print class and header if true */
	protected boolean printHeader() { return true; }

	
	/** Return the max number of fields we can set/get in this message
	 *  Note: that this needs to be defined and executed in the
	 *   final message classes in order to get the right numbers.
	 *  
	 * @return number of message fields
	**/
	public abstract int getNumFields(); // { return NUMFIELDS; }
	
	/** Parse the data buffer message packet into our local fields.
	 *   Prepare received data buffer for access...
	 * @return -- number of bytes parsed (next byte position in buffer)
	 * 			  or -1 on errors
	 */
	protected int parseMessage()
	{
		byte buf[] = getData();
		if( buf == null )	return -1;
		else return ((TMheader)header).parseHeader( buf, 0 );
		// or could add data buffer parsing here...
	}
	
	/** Copy the local field values into a returned int array
	 *  This copies directly from local data buffer,
	 *   interpreting it as defined by getFieldFormat().
	 *   
	 *  Override to use different buffer,
	 *   or more complicated data structures.
	 *   
	 * @return array of fields of length relative to data buffer size
	 *           (nominally NUMFIELDS, but based on actual size)
	 *         or null if no data
	**/
	public int[] getFields()
	{
		byte[] data = getData();
		if( data == null )			return null;
		
		// alloc return array based on size of actual data buffer
		// in case we have variable length messages...
		// otherwise could use:  int size = getNumFields();
		int size = getMsgLen();
		size /= parseHelp.sizeofFMT( getFieldFormat() );
		
		int rvals[] = new int[size];
		int ds = getMsgStart();
		parseHelp.getFields( data, ds,	// source
							 rvals, 0,	// destination
							 fieldFmt,	// array format
							 size );	// number to do
		
		return rvals;
	}
	
	/** Set local fields into our data buffer.
	 *  Prepare local data buffer for send.
	 * @return -- number of bytes set (next byte position in buffer)
	 * 			  or -1 on errors
	 **/
	protected int setMessage()
	{
		allocDataBuffer();
		return ((TMheader)header).setHeader( getData(), 0 );
		// or could add data buffer filling here...
	}
	
	/** Set the local field values from given array.
	 *  This copies the given vals directly into alloced data buffer,
	 *   interpreting them as defined by getFieldFormat().
	 *  Over ride as seen fit.
	 * @return nada
	**/
	public void setFields( int[] vals )
	{
		// get/alloc buffer, set header,
		//  and possibly set local field values in buffer via 
		//  call to setMessage() -- which impl is minimized here
		byte data[] = getDataBuffer();
		
		// figure out how many fields to set
		int size = getMsgLen();
		size /= parseHelp.sizeofFMT( getFieldFormat() );
		if( vals.length < size )	size = vals.length;
			
		int ds = getMsgStart();
		
		parseHelp.setFields( data, ds,	// destination
							 vals, 0,	// source
							 fieldFmt,	// array format
							 size );	// number to do
		
		return;
	}

	/** Utility to set the local field values to the don't-care value
	 *   Possibly 0 or -1...implement as needed...
	 * @return nada
	**/
	public void nullFields() {  }
	
 /**** NOTE: End of Message Parsing and Setting methods. ****/
	
	/** allocate the data buffer if it doesn't already exist
	 */
	protected void allocDataBuffer()
	{
		if( data == null )
			data = ((TMheader)header).allocBuf();
	}
	
	/** Return whatever is the data buffer. For internal use.
	 *  May be null if not yet allocated.
	 */
	protected byte[] getData( ) {  return data;  }
	
	/** Set all fields in local data buffer, then 
	 * @return the full message buffer -- probably to send
	 **/
	public byte[] getDataBuffer()
	{
		setMessage();
		return getData();
	}
	
	/** Set the data buffer and parse it, probably from a recv...
	 **/
	public void setDataBuffer( byte[] b )
	{
		data = b;
		parseMessage();
	}

	/** Utility to set the data buffer to all zeros, if it exists
	 * @return nada
	 *   NOTE: could override to zero our member fields as well...
	**/
	public void zeroDataBuf()
	{
		byte buf[] = getData();
		if( buf != null )
			for( int i=0; i<buf.length; ++i )		buf[i] = 0;
		return;
	}
	
	/** Set the message destination MoteID address
	 *  @param d -- The destination address to set.
	 */
	public void setDestAddr( int d )
	{ ((TMheader)header).setDestAddr( d ); }
	 
	/** Set the message groupID field
	 *  @param g -- The groupID to set.
	 */
	public void setGroupID( int g )
	{ ((TMheader)header).setDestAddr( g ); }
	
	/** Return a user defined message type from the given packet.
	 *  Used by TMfactory methods to make wrapper for received data.
	 *  Uses TMheader definitions to find the type field.
	 * @param b -- raw message packet
	 * @return -- type identifier
	 */
	public int getMsgType( byte[] b )
	{ return ((TMheader)header).getHandlerID( b ); }
	
	/** Return a user defined message type from the message header.
	 *  Uses TMheader definitions to find the type field.
	 * @return -- type identifier
	 */
	public int getMsgType()
	{ return ((TMheader)header).getHandlerID(); }
	
	/** Change message type.
	 *  Probably need to setMsgLen() correctly as well...
	 *   Used in rrDataIO to make a stop command...
	 *   a little sleazy...I lak it...
	 */
	public void setMsgType( int type )
	{ ((TMheader)header).setHandlerID( type ); }
	
	/** Return the user message length.
	 *  Pass through to header definition.
	 * @return -- number of bytes in user message or
	 *            payload size if it is not set.
	 */
	public int getMsgLen()
	{ return ((TMheader)header).getMsgLen(); }
	
	/** Set the message buffer length field,
	 *   for cases of variable length messages
	 *   this should be called before data buffer is
	 *   allocated by calls through to allocDataBuf()...
	 *  Pass through to header definition.
	 *  @param l -- The message buffer length to set.
	 */
	public void setMsgLen( int l )
	{ ((TMheader)header).setMsgLen( l ); }
	
	/** Return the starting byte of the user message payload area.
	 **/
	protected int getMsgStart()
	{ return ((TMheader)header).getPaystart();  }
	
	/** Return a properly typed TMheader class
	 *  TODO need to use generic <T> type again...
	 *   which contains all the system level fields.
	 * @return TMheader++
	 */
	public T getHeader()
	{ return header; }
	
	
	/** Convert message data buffer to String of bytes for printing
	 * NOTE: override and add AM message printing in sub-classes
	 * 		if special formatting is desired
	 */
	public String toHexString()
	{   
		StringBuffer buf = new StringBuffer(200);
		byte data[] = getData();
		if( data != null )
		{
			// format header info
			((TMheader)header).toHexString( buf, data, 0 );
			buf.append( " :: " );
			// format the user data as single bytes
			int i = getMsgStart();
			int e = getMsgLen() + getMsgStart();
			while( i < e )
				buf.append( parseHelp.b2Hex( data[i++] ) + " " );
		}
		
		return buf.toString();
	}
	
	/** Convert message local values to readable format
	 *  Returns contents of user message as hex bytes
	 * NOTE: override and change user message printing in sub-classes
	 **/
	public String toString()
	{
		StringBuffer buf = new StringBuffer(200);
		
		if( printHeader() )
		{
			// put in our class name
			buf.append( this.getClass().getSimpleName() + " -- " );
			// format header info
			buf.append( ((TMheader)header).toString() + " :: " );
		}
		
		// get data values and poot them into String
		int[] vals = getFields();
		if( vals != null )
		{
			// note: use previously defined fieldFmt to do work...
			//pHelp fieldFmt
			
			// number of fields to put on each output line
			int f = getFieldsPerLine();
			// number of fields we have
			int e = vals.length;
			
			int i = 0;		// vals index
			while( i < e )  // while there's anything left
			{
				buf.append("\n ");
				// do fields/line 
				for( int j = 0; (j < f) && (i < e); ++j, ++i )
				{
					buf.append( fieldFmt.toString(vals[i]) + " " );
				}
			}
		}

		return buf.toString();
	}
	
	/** Convert message local values to pretty print format.
	 * NOTE: override and add user message printing in sub-classes
	 **/
	public String toPrettyString()
	{
		return this.toString();
	}
	
	
	/****
	 ***    parse and set support routines
	****/

	/** Set the data payload buffer starting at the specified
	 *   location to the given values.
	 * @param pos -- byte position to set,
	 * 					 from beginning of payload NOT header
	 * @param in -- array of data values
	 * @param num -- number of bytes to copy
	 *  @return -- number of bytes set: num if ok, 0 if out of range
	 */
	public int copyDataBytes( int pos, byte[] in, int num )
	{
		// check range, fail if it overruns
		if( (pos+num) > (getMsgLen()-1) )
			return 0;
		
		// get data buffer (may allocate it too)
		byte[] buf = getDataBuffer();
		System.arraycopy( in, 0, buf, pos+getMsgStart(), num );
		
		return num;
	}
} // end'o'TOS_Msg

/*
class stuff
{
	// FromByte -- Java to TOS
	//  Unsigned convert from a Java int
	//  into a byte in the buffer at 'start'
	public static void fb( int d, byte[] buf, int start )
	{
		buf[start]   = (byte) (d & 0xff);
	}

	// ToByte -- TOS to Java
	// Unsigned convert buffer at 'start'
	//  from a byte to Java int, without sign extension
	public static int tb( byte[] buf, int start )
	{
		return (int) ( buf[start] & 0xff );
	}

	// ToSignedByte -- TOS to Java
	// Signed convert buffer at 'start'
	//  from a byte to Java int, with sign extension
	public static int tbm( byte[] buf, int start )
	{
		return (int) ( buf[start] );
	}

	// FromInt -- Java to TOS
	//  Unsigned convert from a Java int
	//  into a little-*f*ck*ng*-endian int16 in the buffer at byte 'start'
	public static void fi( int d, byte[] buf, int start )
	{
		buf[start]   = (byte) (d & 0xff);
		buf[start+1] = (byte) ((d >> 8) & 0xff);
	}

	// ToUInt -- TOS to Java
	// Unsigned convert buffer at byte 'start'
	//  from a little-*f*ck*ng*-endian int into a Java int, without sign
	// note: this will handle signed ints of less than 2 bytes as well
	public static int ti( byte[] buf, int start )
	{
		return ((buf[start+1] << 8) & 0xff00) + (buf[start] & 0xff);
	}

	// ToSignedInt -- TOS to Java
	// convert 2-byte buffer at byte 'start'
	//  from a little-*f*ck*ng*-endian int into a signed Java int
	// sign extend if necessary
	public static int tim( byte[] buf, int start )
	{
		int x = ((buf[start+1] << 8) & 0xff00) + (buf[start] & 0xff);
		if( (x & 0x8000) != 0 )
		{
			// they were trying to tell us the value is negative...
			// so extend the sign to 4 bytes
			x += 0xffff0000;
		}
		return x;
	}
	
	// b2Hex -- convert given byte value to two char hex string
	 //   adds leading zero if necessary
	public static String b2Hex( int x )
	{
		x &= 0xff;
		return ((x < 0x10) ? "0" : "") + Integer.toHexString(x & 0xff);
	}
	
	// s2Hex -- convert given short value to four char hex string
	 //   adds leading zeros if necessary
	public static String s2Hex( int x )
	{
		x &= 0xffff;
		return( ((x < 0x0010) ? "0" : "") +
				((x < 0x0100) ? "0" : "") +
				((x < 0x1000) ? "0" : "") +
				Integer.toHexString(x & 0xffff) );
	}
	
	// b2Dec -- convert given byte value to unsigned decimal string
	public static String b2Dec( int x )
	{
		return Integer.toString(x & 0xff);
	}
	
	// s2Dec -- convert given short value to unsigned decimal string
	public static String s2Dec( int x )
	{
		return Integer.toString(x & 0xffff);
	}
	 
	// b2Decm -- convert given byte value to signed decimal string
	public static String b2Decm( int x )
	{
		return Integer.toString(x);
	}
	
	// s2Decm -- convert given short value to signed decimal string
	public static String s2Decm( int x )
	{
		return Integer.toString(x);
	}
}
*/