/** Oscope1000Msg.java -- TOS demo Oscilloscope message.
 * 		Uses to CC1000 type message header for mica2 and micaz.
 * 
 *  BUFFER_SIZE = 10
 *	struct OscopeMsg
 *	{
 *	    uint16_t sourceMoteID;
 *	    uint16_t lastSampleNumber;
 *	    uint16_t channel;
 *	    uint16_t data[BUFFER_SIZE];
 *	};
**/
package com.etantdonnes.tinyos.demoMessage;

import com.etantdonnes.tinyos.*;


/** Message received from Mote using CC1000 header type -- mica's
 * This message will be received as a raw buffer from the mote.
 *  The buffer will be stuck into the data area of this message class
 *  and then parsed appropriately.  
 **/
public class Oscope1000Msg extends TMstub1000
{
	// the AM types that match this class
	public static final int AM_OSCOPEMSG = 10;
	public static final int PAYSIZE = 26;	// payload bytes in this packet
	private static final int NUMFIELDS = 13;	// number of parameters in message
	
	// components of the OscopeMsg itself
	public static final int BUFFER_SIZE = 10;
    public int sourceMoteID;
    public int lastSampleNumber;
    public int channel;
    public int samples[];

	/** Create a message with empty values
	 *    usually for received message wrapper
	 */
	public Oscope1000Msg()
	{
		super();
	}

	/** Create a message with specified GroupID and destination
	 * 		usually for send message. Sets type and length correctly.
	 * @param gid	-- group ID
	 * @param dest	-- destination addr
	 */
	public Oscope1000Msg( int gid, int dest )
	{
		super( AM_OSCOPEMSG, PAYSIZE, dest, gid );
	}

	/** Parse a message packet into our local fields.
	 *  @return -- new offset if OK
	 *  		   -1 if something wrong with packet.
	 *
	 * NOTE: all the offsets and sizes are hard coded here...
	 **/
	public int parseMessage()
	{
		// parse header and get the first byte of the payload area
		int s = super.parseMessage();
		if( s < 0 )		return s;
		
		// do the local values
		sourceMoteID		= parseHelp.ti(data, s); s+=2;
    	lastSampleNumber	= parseHelp.ti(data, s); s+=2;
    	channel				= parseHelp.ti(data, s); s+=2;
    	samples = new int[BUFFER_SIZE];
    	for( int i=0; i<samples.length; ++i )
    	{
			samples[i]		= parseHelp.ti(data, s); s+=2;;
    	}

		return s;
	}
	
	/** Set local fields in given data buffer --
	 *   not implemented because we only receive this message type
	 *  @return -- new offset if OK
	 *  		   -1 if something wrong with packet.
	 *
	 * NOTE: all the offsets and sizes are hard coded here...
	 **/
	public int setMessage()
	{
		// set header and get the first byte of the payload area
		int s = super.setMessage();
		if( s < 0 )		return s;

		return s;
	}
	
	/** Set the local values to -1 -- The don't-set value
	 *  not really needed so don't bother
	 * @return nada
	public void nullFields()
	{
		return;
	}
	**/
	
	/** Get the number of fields we can set in this message
	 *  Note that this needs to be defined and executed in
	 *  the final message classes in order to get the right numbers.
	 *  
	 * @return number of parameters
	**/
	public int getNumFields() { return NUMFIELDS; }

	/** convert local message data to readable format
	 **/
	public String toString()
	{
		return( (printHeader() ?
					(// put in our class name
					 this.getClass().getSimpleName() + " -- " +
					 // format header info
					 ((TMheader)header).toString() + "\n" )
					 : ("")) +
				"s=" + sourceMoteID +
		    	" n=" + lastSampleNumber +
		    	" c=" + channel + "\n" +
		    	samples[0] + " " +
		    	samples[1] + " " +
		    	samples[2] + " " +
		    	samples[3] + " " +
		    	samples[4] + " " +
		    	samples[5] + " " +
		    	samples[6] + " " +
		    	samples[7] + " " +
		    	samples[8] + " " +
		    	samples[9]
			  );
	}
	
	/** convert local message data to pretty format
	 **/
	public String toPrettyString()
	{
		// fergit about the header printing....
		return(
				"Source:    " + sourceMoteID + "\n" +
				"Number:    " + lastSampleNumber + "\n" +
				"Channel:   " + channel + "\n" +
				"Samples:    " +
					samples[0] + " " +
			    	samples[1] + " " +
			    	samples[2] + " " +
			    	samples[3] + " " +
			    	samples[4] + " " +
			    	samples[5] + " " +
			    	samples[6] + " " +
			    	samples[7] + " " +
			    	samples[8] + " " +
			    	samples[9]
			  );
	}

} // end'o'Oscope1000Msg
