/** helpParse.java -- helpers for parsing and printing messages.
 *  Moved all the from and to buffer conversions and
 *   string formatters here too, changed from TOS_Msg.
 * @author schip
 */
package com.etantdonnes.tinyos;


public class parseHelp
{
	// possible conversion formats
	public static final int UNSIGNEDBYTE  = 1;
	public static final int SIGNEDBYTE    = 2;
	public static final int UNSIGNEDSHORT = 3;
	public static final int SIGNEDSHORT   = 4;
	
	/** Get the number of bytes for each element of above formats
	 *   for sizing return arrays and printing.
	 * @param fmt -- format selector above
	 * @return -- size of format in bytes
	 */
	static public int sizeofFMT( int fmt )
	{
		if( fmt < UNSIGNEDSHORT )	return 1;
		else						return 2;
	}

	/** Get the [un]signed state for each element of above formats
	 *   for printing.
	 * @param fmt -- format selector above
	 * @return -- false if signed, true if unsigned
	 */
	static public boolean unsignFMT( int fmt )
	{
		if( (fmt == UNSIGNEDBYTE) || (fmt == UNSIGNEDSHORT) )
			return true;
		else
			return false;
	}

	/** Convert fields from given buffer into ints in rvals array
	 *   using the supplied pHelp fmt methods.
	 *  Assumes that all fields are of the same size and type.
	 * 
	 * @param d -- data buffer
	 * @param ds -- starting index of byte in buffer
	 * @param rvals -- return array
	 * @param rs -- starting index of int in rvals
	 * @param fmt -- formatting class specifier, see getFormatter()
	 * @param n -- number of values to convert
	 * @return -- index number of the next byte to convert
	 */
	static public int getFields( byte[] d, int ds,
								 int rvals[], int rs,
								 pHelp fmt, int n )
	{
		// use n as the number counter and rs as the result position
		// increment ds appropriately for type being converted
		// if we blow array chunks, we just return what we got so far
		try
		{
			int s = fmt.sizeofFmt();
			for( ; (n > 0); ++rs, --n )
			{
				rvals[rs] = fmt.fromBuf(d, ds);
				ds += s;
			}
		} catch( Exception e) { ; }
		
		return ds;
	}
	
	/** Convert fields into given buffer from ints in vals array
	 *   using the supplied pHelp fmt methods.
	 *  Assumes that all fields are of the same size and type.
	 *   Note that the vals ints should be sign extended if
	 *    that's the way we want them, so {UN}SIGNED are treated
	 *    the same way no matter...
	 * 
	 * @param d -- data buffer
	 * @param ds -- starting index of byte in buffer
	 * @param vals -- values array
	 * @param rs -- starting index of int in rvals
	 * @param fmt -- format specifier, see getFormatter()
	 * @param n -- number of values to convert
	 * @return -- index number of the next byte to convert
	 */
	static public int setFields( byte[] d, int ds,
								 int vals[], int rs,
								 pHelp fmt, int n )
	{
		
		// use n as the number counter and rs as the vals position
		// increment ds appropriately for type being converted
		// if we blow array chunks, we just return what we got so far
		try
		{
			int s = fmt.sizeofFmt();
			for( ; (n > 0); ++rs, --n )
			{
				fmt.toBuf(vals[rs], d, ds);
				ds += s;
			}
		} catch( Exception e) { ; }
		
		return ds;
	}

	/** return an instance of the right kind of pHelp class
	 *   based on the given arguments.
	 * @param fieldFmt -- field format specs at top of this class
	 * @param printBase -- toString base, 10 or 16 for decimal or hex
	 * @return -- pHelp class with to,from,print methods
	 */
	static public pHelp getFormatter( int fieldFmt, int printBase )
	{
		// size of element (1 or 2 for now)
		int s = parseHelp.sizeofFMT( fieldFmt );
		// signed=false or unsigned=true
		boolean u = parseHelp.unsignFMT( fieldFmt );
		// decimal or hex
		int b = printBase;
		
		// set up to default to unsignedHexByte...
		if( b == 10 )
			if( s == 1 )
				if( u ) // unsigned?
					return new unsignedDecByte();
				else
					return new signedDecByte();
			else // must be 2, no?
				if( u ) // unsigned?
					return new unsignedDecShort();
				else
					return new signedDecShort();
		else // must be 16, no?
			if( s == 2 )
				// if( u ) // we don't care...
					return new unsignedHexShort();
			else // must be 1, no?
				// if( u ) // we don't care...
					return new unsignedHexByte();
		
	}
	
/**** all the raw conversion methods for standalone use ****/

	// FromByte -- Java to TOS
	//  Unsigned convert from a Java int
	//  into a byte in the buffer at 'start'
	public static void fb( int d, byte[] buf, int start )
	{
		buf[start]   = (byte) (d & 0xff);
	}

	// ToByte -- TOS to Java
	// Unsigned convert buffer at 'start'
	//  from a byte to Java int, without sign extension
	public static int tb( byte[] buf, int start )
	{
		return (int) ( buf[start] & 0xff );
	}
	
	// ToSignedByte -- TOS to Java
	// Signed convert buffer at 'start'
	//  from a byte to Java int, with sign extension
	public static int tbm( byte[] buf, int start )
	{
		return (int) ( buf[start] );
	}

	// FromInt -- Java to TOS
	//  Unsigned convert from a Java int
	//  into a little-*f*ck*ng*-endian int16 in the buffer at byte 'start'
	public static void fi( int d, byte[] buf, int start )
	{
		buf[start]   = (byte) (d & 0xff);
		buf[start+1] = (byte) ((d >> 8) & 0xff);
	}

	// ToUInt -- TOS to Java
	// Unsigned convert buffer at byte 'start'
	//  from a little-*f*ck*ng*-endian int into a Java int, without sign
	// note: this will handle signed ints of less than 2 bytes as well
	public static int ti( byte[] buf, int start )
	{
		return ((buf[start+1] << 8) & 0xff00) + (buf[start] & 0xff);
	}

	// ToSignedInt -- TOS to Java
	// convert 2-byte buffer at byte 'start'
	//  from a little-*f*ck*ng*-endian int into a signed Java int
	// sign extend if necessary
	public static int tim( byte[] buf, int start )
	{
		int x = ((buf[start+1] << 8) & 0xff00) + (buf[start] & 0xff);
		if( (x & 0x8000) != 0 )
		{
			// they were trying to tell us the value is negative...
			// so extend the sign to 4 bytes
			x += 0xffff0000;
		}
		return x;
	}
	
	// b2Hex -- convert given byte value to two char hex string
	//   adds leading zero if necessary
	public static String b2Hex( int x )
	{
		x &= 0xff;
		return ((x < 0x10) ? "0" : "") + Integer.toHexString(x & 0xff);
	}
	
	// s2Hex -- convert given short value to four char hex string
	//   adds leading zeros if necessary
	public static String s2Hex( int x )
	{
		x &= 0xffff;
		return( ((x < 0x0010) ? "0" : "") +
				((x < 0x0100) ? "0" : "") +
				((x < 0x1000) ? "0" : "") +
				Integer.toHexString(x & 0xffff) );
	}
	
	// b2Dec -- convert given byte value to unsigned decimal string
	public static String b2Dec( int x )
	{
		return Integer.toString(x & 0xff);
	}
	
	// s2Dec -- convert given short value to unsigned decimal string
	public static String s2Dec( int x )
	{
		return Integer.toString(x & 0xffff);
	}
	
	// b2Decm -- convert given byte value to signed decimal string
	public static String b2Decm( int x )
	{
		return Integer.toString(x);
	}
	
	// s2Decm -- convert given short value to signed decimal string
	public static String s2Decm( int x )
	{
		return Integer.toString(x);
	}

} // end'o'parseHelp

/**** all the pHelp implementations returned by getFormatter,
 ****  they behave as named
 ****/

class unsignedHexByte implements pHelp
{
	public void toBuf( int d, byte[] buf, int pos )
	{ parseHelp.fb( d, buf, pos ); }
	public int fromBuf( byte[] buf, int pos )
	{ return parseHelp.tb( buf, pos ); }
	public int sizeofFmt()
	{ return 1; }
	public String toString( int x )
	{ return parseHelp.b2Hex( x ); }
}

class unsignedHexShort implements pHelp
{
	public void toBuf( int d, byte[] buf, int pos )
	{ parseHelp.fi( d, buf, pos ); }
	public int fromBuf( byte[] buf, int pos )
	{ return parseHelp.ti( buf, pos ); }
	public int sizeofFmt()
	{ return 2; }
	public String toString( int x )
	{ return parseHelp.s2Hex( x ); }
}

class unsignedDecByte implements pHelp
{
	public void toBuf( int d, byte[] buf, int pos )
	{ parseHelp.fb( d, buf, pos ); }
	public int fromBuf( byte[] buf, int pos )
	{ return parseHelp.tb( buf, pos ); }
	public int sizeofFmt()
	{ return 1; }
	public String toString( int x )
	{ return parseHelp.b2Dec( x ); }
}

class signedDecByte implements pHelp
{
	public void toBuf( int d, byte[] buf, int pos )
	{ parseHelp.fb( d, buf, pos ); }
	public int fromBuf( byte[] buf, int pos )
	{ return parseHelp.tb( buf, pos ); }
	public int sizeofFmt()
	{ return 1; }
	public String toString( int x )
	{ return parseHelp.b2Decm( x ); }
}

class unsignedDecShort implements pHelp
{
	public void toBuf( int d, byte[] buf, int pos )
	{ parseHelp.fi( d, buf, pos ); }
	public int fromBuf( byte[] buf, int pos )
	{ return parseHelp.ti( buf, pos ); }
	public int sizeofFmt()
	{ return 2; }
	public String toString( int x )
	{ return parseHelp.s2Dec( x ); }
}

class signedDecShort implements pHelp
{
	public void toBuf( int d, byte[] buf, int pos )
	{ parseHelp.fi( d, buf, pos ); }
	public int fromBuf( byte[] buf, int pos )
	{ return parseHelp.ti( buf, pos ); }
	public int sizeofFmt()
	{ return 2; }
	public String toString( int x )
	{ return parseHelp.s2Decm( x ); }
}
	